
__author__ = 'Niklas Rosenstein <rosensteinniklas@gmail.com>'
__version__ = '1.0.0'

from .enum import Enumeration
from .map import ChainMap, MapAsObject, ObjectAsMap, OrderedDict
from .named import Named
from .record import Record
from .sumtype import Sumtype
from .meta import InlineMetaclassBase
